# -*- coding: utf-8 -*-

# from time import time
import os, glob
from pathlib import Path
from time import time

import networkx as nx
import pandas as pd

import asymmetree.best_matches.LeastResolvedTree as LRT
import asymmetree.tools.GraphTools as gt

from bmgedit.BMGEditing import BMGEditor

# --------------------- parameters ---------------------

bm_estimates = 'Aquificales_best_match_estimates/*.bm_candidates'
result_dir = 'results'
outfile1 = 'Aquificales_statistics.csv'
outfile2 = 'Aquificales_editing.csv'

min_order, max_order = 5, 50        # run editing for components with up to 50 nodes
sim_threshold = 0.8                 # relative threshold for inclusion of subobtimal hits

greedy_repeats = 5

methods = (('mincut', 'cost'),              # cost/gain option irrelevant
           ('bpmf', 'cost'),                # cost/gain option irrelevant
           ('karger', 'cost'), ('karger', 'gain'),
           ('greedy', 'cost'), ('greedy', 'gain'), 
           ('gradient_walk', 'cost'), ('gradient_walk', 'gain'),
           ('louvain', 'cost'),             # cost/gain option irrelevant
           ('louvain_obj', 'cost'), ('louvain_obj', 'gain'))


# --------------------- functions ---------------------

def make_graphs(evalue_cutoff = 1e-5, identity_cutoff = 25, sim = 0.95):
    """Reads the best match estimates from file and returns the weakly
       connected components as digraphs."""
    
    gene_dict = {}
    color_dict = {}
    
    # maps gene id --> list of (hit id, bitscore) tuples
    candidates = {}
    
    full_graph = nx.DiGraph()
    
    # list of resulting connected BMGs
    result = []
    
    for filepath in glob.iglob(bm_estimates):
        print(filepath)
        subject_spec = Path(filepath).stem
        
        if subject_spec not in color_dict:
            color_dict[subject_spec] = len(color_dict)
        
        with open(filepath, 'r') as f:
            for line in f.readlines():
                line = line.split('\t')
                
                if len(line) != 6:
                    continue
                
                query_spec, query_gene, subject_gene, *_ = line[:3]
                evalue = float(line[3])
                bitscore = float(line[4])
                identity = float(line[5])
                
                if query_spec not in color_dict:
                    color_dict[query_spec] = len(color_dict)
                    
                if (query_spec, query_gene) not in gene_dict:
                    q_id = len(gene_dict)
                    gene_dict[(query_spec, query_gene)] = q_id
                    candidates[q_id] = {}
                    full_graph.add_node(q_id, color=color_dict[query_spec])
                    
                if (subject_spec, subject_gene) not in gene_dict:
                    s_id = len(gene_dict)
                    gene_dict[(subject_spec, subject_gene)] = s_id
                    candidates[s_id] = {}
                    full_graph.add_node(s_id, color=color_dict[subject_spec])
                
                q_id = gene_dict[(query_spec, query_gene)]
                s_id = gene_dict[(subject_spec, subject_gene)]
                # q_color = color_dict[query_spec]
                s_color = color_dict[subject_spec]
                if evalue < evalue_cutoff and identity > identity_cutoff:
                    if s_color not in candidates[q_id]:
                        candidates[q_id][s_color] = []
                    candidates[q_id][s_color].append((s_id, bitscore))
    
    for x, colors in candidates.items():
        
        for y_color, y_list in colors.items():
            # sort by bitscore
            y_list.sort(key=lambda item: item[1], reverse=True)
            
            if len(y_list) == 0:
                continue
            else:
                threshold = sim * y_list[0][1]
            for y, bitscore in y_list:
                if bitscore >= threshold:
                    full_graph.add_edge(x, y)
                else:
                    break
    
    for cc in nx.weakly_connected_components(full_graph):
        result.append(full_graph.subgraph(cc).copy())
    
    return result


def write_header(filename):
    
    items = ['sim_ID', 'order', 'size', 'colors', 'dist_is_bmg',
             'method', 'objective',
             'time',
             # 'tp1', 'tn1', 'fp1', 'fn1',   # w.r.t. original BMG
             'inner_nodes1',
             'tp2', 'tn2', 'fp2', 'fn2',   # w.r.t. disturbed graph
             # 'tripleextr_tp1', 'tripleextr_tn1', 'tripleextr_fp1', 'tripleextr_fn1',
             'inner_nodes2',
             'tripleextr_tp2', 'tripleextr_tn2', 'tripleextr_fp2', 'tripleextr_fn2',]
    
    with open(filename, 'w') as f:
        f.write(','.join(items))


def write_line(filename, line):
    
    a, b = 5, 13
    
    with open(filename, 'a') as f:
        
        for i in range(len(methods)):
            f.write('\n')
            f.write( ','.join(str(item) for item in line[:a]) )
            f.write( ',' )
            f.write( ','.join(str(item) for item in line[a+i*b: a+(i+1)*b]) )


def analyze(sim_ID, G):
    
    print(f'sim_ID {sim_ID}')
    
    is_bmg = bool(LRT.is_bmg(G))
    
    colors = set()
    for v in G.nodes():
        colors.add(G.nodes[v]['color'])
    
    data = [sim_ID, G.order(), G.size(), len(colors), is_bmg,]
    
    editor = BMGEditor(G)
        
    for method, objective in methods:
        
        start_time = time()
        editor.build(method, objective=objective)
        runtime = time() - start_time
        
        # without extraction of consistent triples
        bmg1, inner_nodes1 = editor.get_bmg(extract_triples_first=False,
                                            supply_inner_vertex_count=True)
        
        # extract consistent triples and run Aho again
        bmg2, inner_nodes2 = editor.get_bmg(extract_triples_first=True,
                                            supply_inner_vertex_count=True)
        
        data.extend([method, objective,
                     runtime,
                     inner_nodes1,
                     *gt.contingency_table(G, bmg1, as_dict=False),
                     inner_nodes2,
                     *gt.contingency_table(G, bmg2, as_dict=False),
                     ])
    
    return data

    
# ------------------- MAIN -------------------

all_graphs = make_graphs(sim=sim_threshold)

if not os.path.exists(result_dir):
    os.makedirs(result_dir)


# (1) data set statistics
    
df = pd.DataFrame(columns=['order','size', 'colors'])

for i, G in enumerate(all_graphs):
    
    color_set = {G.nodes[v]['color'] for v in G.nodes()}
    
    df = df.append({'order': G.order(),
                    'size': G.size(),
                    'colors': len(color_set)},
                   ignore_index=True)
    
df.to_csv(os.path.join(result_dir, outfile1))


# (2) editing

graphs = [G for G in all_graphs
          if G.order() >= min_order and G.order() <= max_order]

outfile2 = os.path.join(result_dir, outfile2)
write_header(outfile2)

for i, G in enumerate(graphs):
    line = analyze(i, G)
    write_line(outfile2, line)
