# -*- coding: utf-8 -*-

import os, itertools
from time import time

from asymmetree.datastructures.PhyloTree import PhyloTree
from asymmetree.best_matches.TrueBMG import bmg_from_tree
import asymmetree.best_matches.LeastResolvedTree as LRT
import asymmetree.tools.GraphTools as gt

from bmgedit.BMGEditing import BMGEditor

# --------------------- parameters ---------------------

result_dir = 'results'
outfile = 'BMG_edit_comparison.csv'

disturbances = [(0.05, 0.05), (0.1, 0.1), (0.2, 0.2), (0.1, 0.0), (0.0, 0.1)]
orders =            [10, 20, 30 , 40]
color_numbers =     [2, 5, 10, 15, 20]
no_of_graphs =      100
greedy_repeats =    5

methods = (('mincut', 'cost'),              # cost/gain option irrelevant
           ('bpmf', 'cost'),                # cost/gain option irrelevant
           ('karger', 'cost'), ('karger', 'gain'),
           ('greedy', 'cost'), ('greedy', 'gain'), 
           ('gradient_walk', 'cost'), ('gradient_walk', 'gain'),
           ('louvain', 'cost'),             # cost/gain option irrelevant
           ('louvain_obj', 'cost'), ('louvain_obj', 'gain'))


# --------------------- functions ---------------------

def write_header(filename):
    
    items = ['sim_ID', 'repeat', 'ins_prob', 'del_prob', 'N', 'colors',
              'arcs',
              'dist_is_bmg', 'dist_tp', 'dist_tn', 'dist_fp', 'dist_fn',
              
              'method', 'objective',
              'time',
              'tp1', 'tn1', 'fp1', 'fn1',   # w.r.t. original BMG
              'tp2', 'tn2', 'fp2', 'fn2',   # w.r.t. disturbed graph
              'tripleextr_tp1', 'tripleextr_tn1', 'tripleextr_fp1', 'tripleextr_fn1',
              'tripleextr_tp2', 'tripleextr_tn2', 'tripleextr_fp2', 'tripleextr_fn2',]
    
    with open(filename, 'w') as f:
        f.write(','.join(items))


def write_line(filename, line):
    
    a, b = 12, 19
    
    with open(filename, 'a') as f:
        
        for i in range(len(methods)):
            f.write('\n')
            f.write( ','.join(str(item) for item in line[:a]) )
            f.write( ',' )
            f.write( ','.join(str(item) for item in line[a+i*b: a+(i+1)*b]) )


def simulation(sim_ID, repeat, disturbance, N, colors):
    
    print(f'sim_ID {sim_ID}, '\
          f'repeat {repeat+1}/{no_of_graphs}, '\
          f'disturbance (i/d): {disturbance}, '\
          f'N: {N}, '\
          f'colors: {colors}')
    
    random_tree = PhyloTree.random_colored_tree(N, colors,
                                            force_all_colors=True)
    bmg = bmg_from_tree(random_tree)
    disturbed = gt.disturb_graph(bmg, *disturbance,
                                  preserve_properly_colored=True)
    
    is_bmg = bool(LRT.is_bmg(disturbed))
    dist_ct = gt.contingency_table(bmg, disturbed, as_dict=False)
    
    data = [sim_ID, repeat, *disturbance, N, colors,
            bmg.size(),
            is_bmg, *dist_ct,
            ]
    
    editor = BMGEditor(disturbed)
        
    for method, objective in methods:
        
        start_time = time()
        editor.build(method, objective=objective)
        runtime = time() - start_time
        
        # without extraction of consistent triples
        bmg1 = editor.get_bmg(extract_triples_first=False)
        
        # extract consistent triples and run Aho again
        bmg2 = editor.get_bmg(extract_triples_first=True)
        
        data.extend([method, objective,
                     runtime,
                     *gt.contingency_table(bmg, bmg1, as_dict=False),
                     *gt.contingency_table(disturbed, bmg1, as_dict=False),
                     *gt.contingency_table(bmg, bmg2, as_dict=False),
                     *gt.contingency_table(disturbed, bmg2, as_dict=False),
                     ])
    
    return data

    
# ------------------- MAIN -------------------
    
if not os.path.exists(result_dir):
    os.makedirs(result_dir)
    
outfile = os.path.join(result_dir, outfile)

write_header(outfile)

tasks = [(i, *item) for i, item in
         enumerate(t for t in itertools.product(range(no_of_graphs),
                                                disturbances,
                                                orders,
                                                color_numbers)
                   if t[-1] < t[-2])] # less colors than vertices


for task in tasks:
    line = simulation(*task)
    write_line(outfile, line)
