# -*- coding: utf-8 -*-

import os, itertools
import networkx as nx

from asymmetree.datastructures.PhyloTree import PhyloTree
from asymmetree.best_matches.TrueBMG import bmg_from_tree
import asymmetree.best_matches.LeastResolvedTree as LRT
from asymmetree.tools.Build import aho_graph
import asymmetree.tools.GraphTools as gt

from bmgedit.Build import partition
from bmgedit.BMGEditing import get_U1_U2_U3, unsatisfiability_cost

# --------------------- parameters ---------------------

result_dir = 'results'
outfile = 'partition_quality.csv'

disturbances = [(0.05, 0.05), (0.1, 0.1), (0.2, 0.2), (0.1, 0.0), (0.0, 0.1)]
orders =            [10, 20, 30 , 40]
color_numbers =     [2, 5, 10, 15, 20]
no_of_graphs =      200
greedy_repeats =    5

methods = ('mincut',
           'karger',
           'greedy', 'gradient_walk',
           'louvain', 'louvain_obj')


# --------------------- functions ---------------------

def write_header(filename):
    
    items = ['sim_ID', 'repeat', 'ins_prob', 'del_prob', 'N', 'colors',
              'arcs',
              'dist_is_bmg', 'dist_tp', 'dist_tn', 'dist_fp', 'dist_fn',
              'aho_orig_size', 'aho_dist_size',
              'aho_orig_cc', 'aho_dist_cc',
              'aho_orig_min_set', 'aho_dist_min_set',
              'aho_tp', 'aho_tn', 'aho_fp', 'aho_fn',]
    for method in methods:
        for s in ('obj', 'U1', 'U2', 'U3', 'partlen', 'min_set'):
            items.append('{}_{}'.format(method, s))
    
    with open(filename, 'w') as f:
        f.write(','.join(items))


def write_line(filename, line):
    
    with open(filename, 'a') as f:
        f.write('\n')
        f.write( ','.join(str(item) for item in line) )


def simulation(sim_ID, repeat, disturbance, N, colors):
    
    print(f'sim_ID {sim_ID}, '\
          f'repeat {repeat+1}/{no_of_graphs}, '\
          f'disturbance (i/d): {disturbance}, '\
          f'N: {N}, '\
          f'colors: {colors}')
    
    random_tree = PhyloTree.random_colored_tree(N, colors,
                                            force_all_colors=True)
    bmg = bmg_from_tree(random_tree)
    disturbed = gt.disturb_graph(bmg, *disturbance,
                                  preserve_properly_colored=True)
    L = [v for v in bmg.nodes()]
    is_bmg = bool(LRT.is_bmg(disturbed))
    dist_ct = gt.contingency_table(bmg, disturbed, as_dict=False)
    
    R_true, F_true = LRT.informative_forbidden_triples(bmg)
    R_dist, F_dist = LRT.informative_forbidden_triples(disturbed)
    
    aho_true = aho_graph(R_true, L)
    aho_dist = aho_graph(R_dist, L)
    aho_ct = gt.contingency_table(aho_true, aho_dist, as_dict=False)
    
    CC_true = [set(cc) for cc in nx.connected_components(aho_true)]
    CC_dist = [set(cc) for cc in nx.connected_components(aho_dist)]
    aho_true_min_set = min([len(s) for s in CC_true])
    aho_dist_min_set = min([len(s) for s in CC_dist])
    
    data = [sim_ID, repeat, *disturbance, N, colors,
            bmg.size(),
            is_bmg, *dist_ct,
            aho_true.size(), aho_dist.size(),
            len(CC_true), len(CC_dist),
            aho_true_min_set, aho_dist_min_set,
            *aho_ct]
    
    if len(CC_dist) > 1:
        data.extend([float('nan') for _ in range(6 * len(methods))])
    else:
        for method in methods:
            _, part = partition(L, method,
                                obj_function=unsatisfiability_cost,
                                args=(disturbed,),
                                aux_graph=aho_dist,
                                greedy_repeats=greedy_repeats)
            min_set = min([len(s) for s in part])
            U1, U2, U3 = get_U1_U2_U3(part, disturbed)
            data.extend([len(U1) + len(U2) + len(U3),
                         len(U1), len(U2), len(U3),
                         len(part),
                         min_set])
    
    return data


# ------------------- MAIN -------------------
    
if not os.path.exists(result_dir):
    os.makedirs(result_dir)
    
outfile = os.path.join(result_dir, outfile)

write_header(outfile)

tasks = [(i, *item) for i, item in
         enumerate(t for t in itertools.product(range(no_of_graphs),
                                                disturbances,
                                                orders,
                                                color_numbers)
                   if t[-1] < t[-2])] # less colors than vertices


for task in tasks:
    line = simulation(*task)
    write_line(outfile, line)

